@@ -75,11 +75,18 @@ def k8s_auth(func: Callable[..., None]) -> Callable[..., None]:
     :return: The original function wrapped by a function that handles
         k8s API exceptions.
     """
-    try:
-        load_kubernetes_config()
-    except Exception as e:
-        print_warn(f"Could not authenticate with active Kubernetes context. \n--> {e}")
-    return func
+
+    @wraps(func)
+    def wrapper(*args: Any, **kwargs: Any) -> None:
+        try:
+            load_kubernetes_config()
+            return func(*args, **kwargs)
+        except Exception as e:
+            print_warn(
+                f"Could not authenticate with active Kubernetes context. \n--> {e}"
+            )
+
+    return wrapper
 
 
 def handle_k8s_exceptions(func: Callable[..., None]) -> Callable[..., None]:
@@ -159,9 +166,10 @@ def _stage(
     git_url: str = Argument(...),
     git_branch: str = Option("", "--branch"),
     stage_name: str = Argument(...),
+    timeout: int = Argument(None),
 ):
     try:
-        run_stage(stage_name, git_url, git_branch)
+        run_stage(stage_name, git_url, git_branch, timeout=timeout)
         sys.exit(0)
     except Exception:
         sys.exit(1)
@@ -203,15 +211,12 @@ def _create_deployment(
                 characters="=",
                 style="green",
             )
-            with console.status(
-                "[purple]Bodywork deploying[/purple]", spinner="aesthetic"
-            ):
-                run_workflow(
-                    git_url,
-                    git_branch,
-                    ssh_key_path=ssh_key_path,
-                    docker_image_override=image,
-                )
+            run_workflow(
+                git_url,
+                git_branch,
+                ssh_key_path=ssh_key_path,
+                docker_image_override=image,
+            )
             console.rule(characters="=", style="green")
         except BodyworkWorkflowExecutionError:
             sys.exit(1)
@@ -266,6 +271,7 @@ def _update_deployment(
     git_branch: str = Option("", "--branch"),
     asynchronous: bool = Option(False, "--async", hidden=True),
     asynchronous_job_name: str = Option("", "--async-job-name", hidden=True),
+    ssh_key_path: str = Option("", "--ssh"),
     image: str = Option(None, "--bodywork-image", hidden=True),
     retries: int = Option(1),
 ):
@@ -274,7 +280,7 @@ def _update_deployment(
         git_branch=git_branch,
         asynchronous=asynchronous,
         asynchronous_job_name=asynchronous_job_name,
-        ssh_key_path="",
+        ssh_key_path=ssh_key_path,
         secrets_group="",
         image=image,
         retries=retries,
@@ -285,7 +291,8 @@ def _update_deployment(
 @handle_k8s_exceptions
 @k8s_auth
 def _delete_deployment(
-    name: str = Argument(...), asynchronous: bool = Option(False, "--async", hidden=True)
+    name: str = Argument(...),
+    asynchronous: bool = Option(False, "--async", hidden=True),
 ):
     if asynchronous:
         delete_workflow_job(BODYWORK_NAMESPACE, name)
@@ -395,9 +402,7 @@ def _create_secret(
 @get.command("secrets")
 @handle_k8s_exceptions
 @k8s_auth
-def _get_secret(
-    name: str = Argument(None), group: str = Option(None)
-):
+def _get_secret(name: str = Argument(None), group: str = Option(None)):
     if name and not group:
         print_warn("Please specify which secrets group the secret belongs to.")
         sys.exit(1)
@@ -431,9 +436,7 @@ def _update_secret(
 @delete.command("secret")
 @handle_k8s_exceptions
 @k8s_auth
-def _delete_secret(
-    name: str = Argument(None), group: str = Option(None)
-):
+def _delete_secret(name: str = Argument(None), group: str = Option(None)):
     if name:
         if not group:
             print_warn("Please specify which secrets group the secret belongs to.")
