@@ -1,5 +1,5 @@
 # bodywork - MLOps on Kubernetes.
-# Copyright (C) 2020-2021  Bodywork Machine Learning Ltd.
+# Copyright (C) 2020-2022  Bodywork Machine Learning Ltd.
 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Affero General Public License as published
@@ -26,14 +26,15 @@ import yaml
 
 from .constants import BODYWORK_CONFIG_VERSION
 from .exceptions import (
+    BodyworkConfigFileExistsError,
     BodyworkConfigParsingError,
     BodyworkConfigMissingSectionError,
     BodyworkConfigVersionMismatchError,
-    BodyworkConfigValidationError
+    BodyworkConfigValidationError,
 )
 
 DAG = Iterable[Iterable[str]]
-VALID_K8S_NAME_REGEX = r'[a-zA-Z0-9-\.]+'
+VALID_K8S_NAME_REGEX = r"[a-zA-Z0-9-\.]+"
 
 
 class DictDataValidator:
@@ -51,7 +52,7 @@ class DictDataValidator:
         """
         self._data_validator = cerberus.Validator(schema=schema, allow_unknown=True)
 
-    def find_errors_in(self, data: Dict[str, Any], prefix: str = '') -> List[str]:
+    def find_errors_in(self, data: Dict[str, Any], prefix: str = "") -> List[str]:
         """Find schema invalidation errors.
 
         :param data: Data to validate against the schema.
@@ -67,7 +68,7 @@ class DictDataValidator:
             return self._format_errors(errors, prefix)
 
     @staticmethod
-    def _format_errors(errors: Dict[str, Any], prefix: str = '') -> List[str]:
+    def _format_errors(errors: Dict[str, Any], prefix: str = "") -> List[str]:
         """Generate human-readable error messages.
 
         :param errors: Raw error output from data validation.
@@ -75,11 +76,12 @@ class DictDataValidator:
             error, defaults to ''.
         :return: List of formatted errors.
         """
+
         def format_error(k: str, v: Any) -> str:
             try:
                 err_msg = f'{prefix}{k} -> {", ".join(v)}'
             except TypeError:
-                err_msg = f'{prefix}{k} -> {json.dumps(v)}'
+                err_msg = f"{prefix}{k} -> {json.dumps(v)}"
             return err_msg
 
         return [format_error(k, v) for k, v in errors.items()]
@@ -99,7 +101,7 @@ class BodyworkConfig:
             parsed as valid YAML.
         """
         try:
-            config_yaml = config_file_path.read_text(encoding='utf-8', errors='strict')
+            config_yaml = config_file_path.read_text(encoding="utf-8", errors="strict")
             config = yaml.load(config_yaml, Loader=yaml.SafeLoader)
             if type(config) is not dict:
                 raise yaml.YAMLError
@@ -107,7 +109,7 @@ class BodyworkConfig:
             self._config_file_path = config_file_path
             self._root_dir = config_file_path.parent
         except (FileNotFoundError, IsADirectoryError):
-            raise FileExistsError(f'no config file found at {config_file_path}')
+            raise BodyworkConfigFileExistsError(config_file_path)
         except yaml.YAMLError as e:
             raise BodyworkConfigParsingError(config_file_path) from e
         self.check_py_modules_exist = check_py_modules_exist
@@ -116,88 +118,88 @@ class BodyworkConfig:
     def _validate_parsed_config(self) -> None:
         """Validate configuration parameters.
 
-        This function exists seperately to the class constructor purely
+        This function exists separately to the class constructor purely
         to facilitate easier testing.
 
         :raises BodyworkConfigMissingSectionError: if config file does
             not contain all of the following sections: version,
-            project, stages and logging.
+            pipeline, stages and logging.
         :raises BodyworkConfigVersionMismatchError: if config file
-            schema version does not match the schame version supported
+            schema version does not match the schema version supported
             by the current Bodywork version.
         :raises BodyworkConfigValidationError: if a config
             parameter is missing or has been set to an invalid value.
         """
         config = self._config
         missing_config_sections = []
-        if 'version' not in config:
-            missing_config_sections.append('version')
-        if 'project' not in config:
-            missing_config_sections.append('project')
-        if 'stages' not in config:
-            missing_config_sections.append('stages')
-        if 'logging' not in config:
-            missing_config_sections.append('logging')
+        if "version" not in config:
+            missing_config_sections.append("version")
+        if "pipeline" not in config:
+            missing_config_sections.append("pipeline")
+        if "stages" not in config:
+            missing_config_sections.append("stages")
+        if "logging" not in config:
+            missing_config_sections.append("logging")
         if missing_config_sections:
             raise BodyworkConfigMissingSectionError(missing_config_sections)
 
         try:
-            if len(config['version'].split('.')) != 2:
+            if len(config["version"].split(".")) != 2:
                 raise ValueError
-            if config['version'] != BODYWORK_CONFIG_VERSION:
-                raise BodyworkConfigVersionMismatchError(config['version'])
+            if config["version"] != BODYWORK_CONFIG_VERSION:
+                raise BodyworkConfigVersionMismatchError(config["version"])
         except (AttributeError, ValueError):
-            raise BodyworkConfigValidationError(['version'])
+            raise BodyworkConfigValidationError(["version"])
 
         missing_or_invalid_param: List[str] = []
         try:
-            self.project = ProjectConfig(config['project'])
+            self.pipeline = PipelineConfig(config["pipeline"])
         except BodyworkConfigValidationError as e:
             missing_or_invalid_param += e.missing_params
 
         try:
-            self.logging = LoggingConfig(config['logging'])
+            self.logging = LoggingConfig(config["logging"])
         except BodyworkConfigValidationError as e:
             missing_or_invalid_param += e.missing_params
 
         try:
             self.stages: Dict[str, StageConfig] = {}
-            for stage_name, stage_config in config['stages'].items():
-                if 'batch' in stage_config and 'service' in stage_config:
+            for stage_name, stage_config in config["stages"].items():
+                if "batch" in stage_config and "service" in stage_config:
                     missing_or_invalid_param.append(
-                        f'stages.{stage_name}.batch/service'
+                        f"stages.{stage_name}.batch/service"
                     )
                     continue
-                elif 'batch' in stage_config:
+                elif "batch" in stage_config:
                     self.stages[stage_name] = BatchStageConfig(
-                        str(stage_name),
-                        stage_config,
-                        self._root_dir
+                        str(stage_name), stage_config, self._root_dir
                     )
-                elif 'service' in stage_config:
+                elif "service" in stage_config:
                     self.stages[stage_name] = ServiceStageConfig(
-                        str(stage_name),
-                        stage_config,
-                        self._root_dir
+                        str(stage_name), stage_config, self._root_dir
                     )
                 else:
                     missing_or_invalid_param.append(
-                        f'stages.{stage_name}.batch/service'
+                        f"stages.{stage_name}.batch/service"
                     )
         except AttributeError:
-            missing_or_invalid_param.append('stages._ - no stage configs provided')
+            missing_or_invalid_param.append("stages._ - no stage configs provided")
 
         stages_in_workflow_without_valid_config = _check_workflow_stages_are_configured(
-            self.project.workflow,
-            self.stages.keys()
+            self.pipeline.workflow, self.stages.keys()
         )
         missing_or_invalid_param += stages_in_workflow_without_valid_config
-
+        if self.pipeline.run_on_failure:
+            if self.pipeline.run_on_failure not in self.stages.keys():
+                missing_or_invalid_param.append(
+                    f"pipeline.run_on_failure -> cannot find valid stage: "
+                    f"{self.pipeline.run_on_failure} to run on workflow failure."
+                )
         if self.check_py_modules_exist:
             for stage_name, stage in self.stages.items():
                 if not stage.executable_module_path.exists():
                     missing_or_invalid_param.append(
-                        f'stages.{stage_name}.executable_module_path -> does not exist'
+                        f"stages.{stage_name}.executable_module_path -> does not exist"
                     )
 
         if missing_or_invalid_param:
@@ -205,124 +207,113 @@ class BodyworkConfig:
             raise BodyworkConfigValidationError(missing_or_invalid_param)
 
 
-class ProjectConfig:
-    """High-level project configuration."""
+class PipelineConfig:
+    """High-level pipeline configuration."""
 
     SCHEMA = {
-        'name': {
-            'type': 'string',
-            'required': True,
-            'regex': VALID_K8S_NAME_REGEX
-        },
-        'docker_image': {
-            'type': 'string',
-            'required': True
-        },
-        'DAG': {
-            'type': 'string',
-            'required': True
-        }
+        "name": {"type": "string", "required": True, "regex": VALID_K8S_NAME_REGEX},
+        "docker_image": {"type": "string", "required": True},
+        "DAG": {"type": "string", "required": True},
+        "usage_stats": {"type": "boolean", "required": False},
+        "run_on_failure": {"type": "string", "required": False},
+        "namespace": {"type": "string", "required": False},
+        "secrets_group": {"type": "string", "required": False},
     }
 
     def __init__(self, config_section: Dict[str, str]):
         """Constructor.
 
-        :param config: Dictionary of configuration parameters.
+        :param config_section: Dictionary of configuration parameters.
         :raises BodyworkConfigValidationError: if any
             required configuration parameters are missing or invalid.
         """
 
         data_validator = DictDataValidator(self.SCHEMA)
         missing_or_invalid_param = data_validator.find_errors_in(
-            config_section,
-            prefix='project.'
+            config_section, prefix="pipeline."
         )
         if missing_or_invalid_param:
             raise BodyworkConfigValidationError(missing_or_invalid_param)
         else:
-            self.name = config_section['name']
-            self.docker_image = config_section['docker_image']
-            self.DAG = config_section['DAG']
+            self.name = config_section["name"]
+            self.docker_image = config_section["docker_image"]
+            self.DAG = config_section["DAG"]
+            self.usage_stats = (
+                config_section["usage_stats"]
+                if "usage_stats" in config_section
+                else True
+            )
+            self.run_on_failure = (
+                config_section["run_on_failure"]
+                if "run_on_failure" in config_section
+                else ""
+            )
+            self.namespace = (
+                config_section["namespace"] if "namespace" in config_section else ""
+            )
+            self.secrets_group = (
+                config_section["secrets_group"]
+                if "secrets_group" in config_section
+                else ""
+            )
             try:
-                self.workflow = _parse_dag_definition(config_section['DAG'])
+                self.workflow = _parse_dag_definition(config_section["DAG"])
             except ValueError as e:
-                raise BodyworkConfigValidationError([f'project.DAG -> {e}'])
+                raise BodyworkConfigValidationError([f"pipeline.DAG -> {e}"])
 
 
 class LoggingConfig:
     """Logging configuration."""
 
     SCHEMA = {
-        'log_level': {
-            'type': 'string',
-            'required': True,
-            'allowed': ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
+        "log_level": {
+            "type": "string",
+            "required": True,
+            "allowed": ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
         }
     }
 
     def __init__(self, config_section: Dict[str, str]):
         """Constructor.
 
-        :param config: Dictionary of configuration parameters.
+        :param config_section: Dictionary of configuration parameters.
         :raises BodyworkConfigValidationError: if any
             required configuration parameters are missing or invalid.
         """
 
         data_validator = DictDataValidator(self.SCHEMA)
         missing_or_invalid_param = data_validator.find_errors_in(
-            config_section,
-            prefix='logging.'
+            config_section, prefix="logging."
         )
         if missing_or_invalid_param:
             raise BodyworkConfigValidationError(missing_or_invalid_param)
         else:
-            self.log_level = config_section['log_level']
+            self.log_level = config_section["log_level"]
 
 
 class StageConfig:
     """Common stage configuration for all stages."""
 
     SCHEMA_GENERIC = {
-        'executable_module_path': {
-            'type': 'string',
-            'required': True,
-            'regex': r'.+(\.py$)'
+        "executable_module_path": {
+            "type": "string",
+            "required": True,
+            "regex": r".+\.(py|ipynb)$",
         },
-        'args': {
-            'type': 'list',
-            'required': False,
-            'schema': {
-                'type': 'string'
-            }
+        "args": {"type": "list", "required": False, "schema": {"type": "string"}},
+        "cpu_request": {"type": "float", "required": True, "min": 0.0},
+        "memory_request_mb": {"type": "integer", "required": True, "min": 0},
+        "requirements": {
+            "type": "list",
+            "required": False,
+            "schema": {"type": "string"},
         },
-        'cpu_request': {
-            'type': 'float',
-            'required': True,
-            'min': 0.0
+        "secrets": {
+            "type": "dict",
+            "required": False,
+            "keysrules": {"type": "string"},
+            "valuesrules": {"type": "string", "regex": VALID_K8S_NAME_REGEX},
         },
-        'memory_request_mb': {
-            'type': 'integer',
-            'required': True,
-            'min': 0
-        },
-        'requirements': {
-            'type': 'list',
-            'required': False,
-            'schema': {
-                'type': 'string'
-            }
-        },
-        'secrets': {
-            'type': 'dict',
-            'required': False,
-            'keysrules': {
-                'type': 'string'
-            },
-            'valuesrules': {
-                'type': 'string',
-                'regex': VALID_K8S_NAME_REGEX
-            }
-        }
     }
 
     def __init__(self, stage_name: str, config: Dict[str, Any], root_dir: Path):
@@ -330,31 +321,32 @@ class StageConfig:
 
         :param stage_name: Name of stage.
         :param config: Dictionary of configuration parameters.
-        :param root_dir: The root directory of the project containing
+        :param root_dir: The root directory of the pipeline containing
             the bodywork config file and the stage directories.
         """
         data_validator = DictDataValidator(self.SCHEMA_GENERIC)
         self._missing_or_invalid_param = data_validator.find_errors_in(
-            config,
-            prefix=f'stages.{stage_name}.'
+            config, prefix=f"stages.{stage_name}."
         )
         if not self._missing_or_invalid_param:
             self.name = stage_name
-            self.executable_module_path = root_dir / config['executable_module_path']
+            self.executable_module_path = root_dir / config["executable_module_path"]
             self.executable_module = self.executable_module_path.name
-            self.args = config['args'] if 'args' in config else []
-            self.cpu_request = config['cpu_request']
-            self.memory_request = config['memory_request_mb']
-            self.requirements = config['requirements'] if 'requirements' in config else []  # noqa
-            if 'secrets' in config:
+            self.args = config["args"] if "args" in config else []
+            self.cpu_request = config["cpu_request"]
+            self.memory_request = config["memory_request_mb"]
+            self.requirements = (
+                config["requirements"] if "requirements" in config else []
+            )
+            if "secrets" in config:
                 self.env_vars_from_secrets = [
                     (secret_name, secret_key)
-                    for secret_key, secret_name in config['secrets'].items()
+                    for secret_key, secret_name in config["secrets"].items()
                 ]
             else:
                 self.env_vars_from_secrets = []
 
-    def __eq__(self, other) -> bool:
+    def __eq__(self, other: Any) -> bool:
         """Object equality operator.
 
         :param other: Other Stage object to compare this one too.
@@ -369,16 +361,8 @@ class BatchStageConfig(StageConfig):
     """Specific stage configuration for batch stages."""
 
     SCHEMA_BATCH = {
-        'max_completion_time_seconds': {
-            'type': 'integer',
-            'required': True,
-            'min': 0
-        },
-        'retries': {
-            'type': 'integer',
-            'required': True,
-            'min': 0
-        }
+        "max_completion_time_seconds": {"type": "integer", "required": True, "min": 0},
+        "retries": {"type": "integer", "required": True, "min": 0},
     }
 
     def __init__(self, stage_name: str, config: Dict[str, Any], root_dir: Path):
@@ -386,21 +370,20 @@ class BatchStageConfig(StageConfig):
 
         :param stage_name: Name of parent stage config.
         :param config: Dictionary of configuration parameters.
-        :param root_dir: The root directory of the project containing
+        :param root_dir: The root directory of the pipeline containing
             the bodywork config file and the stage directories.
         :raises BodyworkConfigValidationError: if any
             required configuration parameters are missing or invalid.
         """
         super().__init__(stage_name, config, root_dir)
-        batch_config = config['batch']
+        batch_config = config["batch"]
         data_validator = DictDataValidator(self.SCHEMA_BATCH)
         self._missing_or_invalid_param += data_validator.find_errors_in(
-            batch_config,
-            prefix=f'stages.{stage_name}.batch.'
+            batch_config, prefix=f"stages.{stage_name}.batch."
         )
         if not self._missing_or_invalid_param:
-            self.max_completion_time = batch_config['max_completion_time_seconds']
-            self.retries = batch_config['retries']
+            self.max_completion_time = batch_config["max_completion_time_seconds"]
+            self.retries = batch_config["retries"]
         else:
             raise BodyworkConfigValidationError(self._missing_or_invalid_param)
 
@@ -409,49 +392,36 @@ class ServiceStageConfig(StageConfig):
     """Specific stage configuration for service stages."""
 
     SCHEMA_SERVICE = {
-        'max_startup_time_seconds': {
-            'type': 'integer',
-            'required': True,
-            'min': 0
+        "max_startup_time_seconds": {"type": "integer", "required": True, "min": 0},
+        "replicas": {"type": "integer", "required": True, "min": 0},
+        "port": {"type": "integer", "required": True, "min": 0},
+        "ingress": {
+            "type": "boolean",
+            "required": True,
         },
-        'replicas': {
-            'type': 'integer',
-            'required': True,
-            'min': 0
-        },
-        'port': {
-            'type': 'integer',
-            'required': True,
-            'min': 0
-        },
-        'ingress': {
-            'type': 'boolean',
-            'required': True,
-        }
     }
 
-    def __init__(self, stage_name, config: Dict[str, Any], root_dir: Path):
+    def __init__(self, stage_name: str, config: Dict[str, Any], root_dir: Path) -> None:
         """Constructor.
 
         :param stage_name: Name of parent stage config.
         :param config: Dictionary of configuration parameters.
-        :param root_dir: The root directory of the project containing
+        :param root_dir: The root directory of the pipeline containing
             the bodywork config file and the stage directories.
         :raises BodyworkConfigValidationError: if any
             required configuration parameters are missing or invalid.
         """
         super().__init__(stage_name, config, root_dir)
-        service_config = config['service']
+        service_config = config["service"]
         data_validator = DictDataValidator(self.SCHEMA_SERVICE)
         self._missing_or_invalid_param += data_validator.find_errors_in(
-            service_config,
-            prefix=f'stages.{stage_name}.service.'
+            service_config, prefix=f"stages.{stage_name}.service."
         )
         if not self._missing_or_invalid_param:
-            self.max_startup_time = service_config['max_startup_time_seconds']
-            self.replicas = service_config['replicas']
-            self.port = service_config['port']
-            self.create_ingress = service_config['ingress']
+            self.max_startup_time = service_config["max_startup_time_seconds"]
+            self.replicas = service_config["replicas"]
+            self.port = service_config["port"]
+            self.create_ingress = service_config["ingress"]
         else:
             raise BodyworkConfigValidationError(self._missing_or_invalid_param)
 
@@ -463,36 +433,38 @@ def _parse_dag_definition(dag_definition: str) -> DAG:
     :raises ValueError: If any 'null' (zero character) stage names are
         found.
     :return: A list of steps, where each step is a list of Bodywork
-        project stage names (containing a list of stages to run in each
+        pipeline stage names (containing a list of stages to run in each
         step).
     """
-    steps = dag_definition.replace(' ', '').split('>>')
-    stages_in_steps = [step.split(',') for step in steps]
+    steps = dag_definition.replace(" ", "").split(">>")
+    stages_in_steps = [step.split(",") for step in steps]
     steps_with_null_stages = [
         str(n)
-        for n, step in enumerate(stages_in_steps, start=1) for stage in step
-        if stage == ''
+        for n, step in enumerate(stages_in_steps, start=1)
+        for stage in step
+        if stage == ""
     ]
     if len(steps_with_null_stages) > 0:
-        msg = (f'null stages found in step {", ".join(steps_with_null_stages)} when '
-               f'parsing DAG definition')
+        msg = (
+            f'null stages found in step {", ".join(steps_with_null_stages)} when '
+            f"parsing DAG definition"
+        )
         raise ValueError(msg)
     return stages_in_steps
 
 
 def _check_workflow_stages_are_configured(
-    workflow: Iterable[Iterable[str]],
-    stages: Iterable[str]
+    workflow: Iterable[Iterable[str]], stages: Iterable[str]
 ) -> Sequence[str]:
     """Identify stages in workflow that have not been configured.
 
-    :param workflow: A project DAG parsed into a Bodywork workflow.
+    :param workflow: A pipeline DAG parsed into a Bodywork workflow.
     :param stages: List of stages that have been configured.
     :return: List of missing stage messages.
     """
     stages_in_workflow = [stage for step in workflow for stage in step]
     missing_stages = [
-        f'project.workflow -> cannot find valid stage @ stages.{stage}'
+        f"pipeline.workflow -> cannot find valid stage @ stages.{stage}"
         for stage in stages_in_workflow
         if stage not in stages
     ]
