@@ -32,6 +32,7 @@ BODYWORK_WORKFLOW_SERVICE_ACCOUNT = "bodywork-workflow-controller"
 BODYWORK_WORKFLOW_JOB_TIME_TO_LIVE = 15 * 60
 BODYWORK_STAGES_SERVICE_ACCOUNT = "bodywork-stage"
 BODYWORK_NAMESPACE = "bodywork-deployment-jobs"
+DEFAULT_K8S_POLLING_FREQ = 2
 DEFAULT_LOG_LEVEL = "INFO"
 DEFAULT_LOG_LEVEL_ENV_VAR = "BODYWORK_LOG_LEVEL"
 DEFAULT_PROJECT_DIR = Path("./bodywork_project")
@@ -41,12 +42,16 @@ DEFAULT_SSH_FILE = "id_bodywork"
 FAILURE_EXCEPTION_K8S_ENV_VAR = "EXCEPTION_MESSAGE"
 GIT_SSH_COMMAND = "GIT_SSH_COMMAND"
 GIT_COMMIT_HASH_K8S_ENV_VAR = "GIT_COMMIT_HASH"
+K8S_MAX_SURGE = 2
+K8S_MAX_UNAVAILABLE = 0
+K8S_PROBE_PERIOD_SECONDS = 10
+LOG_TIME_FORMAT = "[%d/%m/%y %H:%M:%S]"
 PROJECT_CONFIG_FILENAME = "bodywork.yaml"
 SSH_DIR_NAME = ".ssh"
 SECRET_GROUP_LABEL = "group"
 SSH_PRIVATE_KEY_ENV_VAR = "BODYWORK_GIT_SSH_PRIVATE_KEY"
 SSH_SECRET_NAME = "ssh-git-private-key"
-TIMEOUT_GRACE_SECONDS = 90
+TIMEOUT_GRACE_SECONDS = 30
 USAGE_STATS_SERVER_URL = "http://k8s.bodyworkml-dev.com/bodywork-ml/usage-tracking--server/workflow-execution-counter"  # noqa
 
 # External SSH Fingerprints
