@@ -62,7 +62,7 @@ from .secrets import (
     create_ssh_key_secret_from_file,
     create_secret_env_variable,
     delete_secret_group,
-    secret_group_exists
+    secret_group_exists,
 )
 from .deployments import (
     DeploymentStatus,
@@ -89,6 +89,7 @@ from .utils import (
     api_exception_msg,
     create_k8s_environment_variables,
     EnvVars,
+    has_unscheduleable_pods,
     make_valid_k8s_name,
 )
 
@@ -146,6 +147,7 @@ __all__ = [
     "api_exception_msg",
     "create_k8s_environment_variables",
     "EnvVars",
+    "has_unscheduleable_pods",
     "make_valid_k8s_name",
     "replicate_secrets_in_namespace",
     "update_secret",
@@ -155,5 +157,5 @@ __all__ = [
     "create_ssh_key_secret_from_file",
     "create_secret_env_variable",
     "delete_secret_group",
-    "secret_group_exists"
+    "secret_group_exists",
 ]
