@@ -18,9 +18,11 @@
 This module contains all of the functions required to execute and manage
 a Bodywork project workflow - a sequence of stages represented as a DAG.
 """
+from datetime import datetime, timedelta
+from math import ceil
 from pathlib import Path
 from shutil import rmtree
-from typing import cast, Tuple, List, Any
+from typing import Any, cast, List, Tuple
 from kubernetes.client.exceptions import ApiException
 
 import requests
@@ -28,13 +30,15 @@ import os
 import stat
 
 from . import k8s
-from .cli.terminal import print_pod_logs
+from .cli.terminal import make_progress_bar, print_pod_logs
 from .config import BodyworkConfig, BatchStageConfig, ServiceStageConfig
 from .constants import (
     DEFAULT_PROJECT_DIR,
     PROJECT_CONFIG_FILENAME,
     TIMEOUT_GRACE_SECONDS,
     GIT_COMMIT_HASH_K8S_ENV_VAR,
+    K8S_MAX_SURGE,
+    K8S_MAX_UNAVAILABLE,
     USAGE_STATS_SERVER_URL,
     FAILURE_EXCEPTION_K8S_ENV_VAR,
     BODYWORK_STAGES_SERVICE_ACCOUNT,
@@ -42,6 +46,7 @@ from .constants import (
     SSH_SECRET_NAME,
 )
 from .exceptions import (
+    BodyworkJobFailure,
     BodyworkWorkflowExecutionError,
     BodyworkNamespaceError,
     BodyworkDockerImageError,
@@ -122,7 +127,7 @@ def run_workflow(
             _copy_secrets_to_target_namespace(namespace, config.pipeline.name)
 
         for step in workflow_dag:
-            _log.info(f"Attempting to execute DAG step = [{', '.join(step)}]")
+            _log.info(f"Executing DAG step = [{', '.join(step)}]")
             batch_stages = [
                 cast(BatchStageConfig, all_stages[stage_name])
                 for stage_name in step
@@ -286,6 +291,7 @@ def _run_batch_stages(
             repo_url,
             repo_branch,
             retries=stage.retries,
+            timeout=stage.max_completion_time,
             container_env_vars=k8s.configure_env_vars_from_secrets(
                 namespace, stage.env_vars_from_secrets
             )
@@ -301,16 +307,28 @@ def _run_batch_stages(
         _log.info(f"Creating k8s job for stage = {job_name}")
         k8s.create_job(job_object)
     try:
-        timeout = max(stage.max_completion_time for stage in batch_stages)
-        k8s.monitor_jobs_to_completion(job_objects, timeout + TIMEOUT_GRACE_SECONDS)
-    except TimeoutError as e:
+        timeout = _compute_optimal_job_timeout(batch_stages)
+        timeout_dt = (datetime.now() + timedelta(seconds=timeout)).strftime(
+            "%d/%m/%y %H:%M:%S"
+        )
+        _log.info(f"Monitoring k8s jobs, timeout at [{timeout_dt}] ({timeout}s)")
+        with make_progress_bar(timeout) as progress_bar:
+            k8s.monitor_jobs_to_completion(
+                job_objects, timeout, progress_bar=progress_bar
+            )
+        for job_object in job_objects:
+            job_name = job_object.metadata.name
+            _log.info(f"Completed k8s job for stage = {job_name}")
+            _print_logs_to_stdout(namespace, job_name, False)
+    except (TimeoutError, BodyworkJobFailure) as e:
         _log.error("Some (or all) k8s jobs failed to complete successfully")
+        for job_object in job_objects:
+            job_name = job_object.metadata.name
+            _print_logs_to_stdout(namespace, job_name, True)
         raise e
     finally:
         for job_object in job_objects:
             job_name = job_object.metadata.name
-            _log.info(f"Completed k8s job for stage = {job_name}")
-            _print_logs_to_stdout(namespace, job_name)
             _log.info(f"Deleting k8s job for stage = {job_name}")
             k8s.delete_job(namespace, job_name)
             _log.info(f"Deleted k8s job for stage = {job_name}")
@@ -354,7 +372,7 @@ def _run_service_stages(
             image=docker_image,
             cpu_request=stage.cpu_request,
             memory_request=stage.memory_request,
-            seconds_to_be_ready_before_completing=stage.max_startup_time,
+            startup_time_seconds=stage.max_startup_time,
         )
         for stage in service_stages
     ]
@@ -369,15 +387,20 @@ def _run_service_stages(
             )
             k8s.create_deployment(deployment_object)
     try:
-        timeout = max(stage.max_startup_time for stage in service_stages)
-        k8s.monitor_deployments_to_completion(
-            deployment_objects, timeout + TIMEOUT_GRACE_SECONDS
+        timeout = _compute_optimal_deployment_timeout(namespace, service_stages)
+        timeout_dt = (datetime.now() + timedelta(seconds=timeout)).strftime(
+            "%d/%m/%y %H:%M:%S"
         )
+        _log.info(f"Monitoring k8s deployments, timeout at [{timeout_dt}] ({timeout}s)")
+        with make_progress_bar(timeout) as progress_bar:
+            k8s.monitor_deployments_to_completion(
+                deployment_objects, timeout, progress_bar=progress_bar
+            )
     except TimeoutError as e:
         _log.error("Deployments failed to roll-out successfully")
         for deployment_object in deployment_objects:
             deployment_name = deployment_object.metadata.name
-            _print_logs_to_stdout(namespace, deployment_name)
+            _print_logs_to_stdout(namespace, deployment_name, True)
             _log.info(f"Rolling-back k8s deployment for stage = {deployment_name}")
             k8s.rollback_deployment(deployment_object)
             _log.info(f"Rolled-back k8s deployment for stage = {deployment_name}")
@@ -387,7 +410,7 @@ def _run_service_stages(
         deployment_name = deployment_object.metadata.name
         deployment_port = deployment_object.metadata.annotations["port"]
         _log.info(f"Successfully created k8s deployment for stage = {deployment_name}")
-        _print_logs_to_stdout(namespace, deployment_name)
+        _print_logs_to_stdout(namespace, deployment_name, False)
         if not k8s.is_exposed_as_cluster_service(namespace, deployment_name):
             _log.info(
                 f"Exposing stage = {deployment_name} as a k8s service at "
@@ -490,16 +513,63 @@ def parse_dockerhub_image_string(image_string: str) -> Tuple[str, str]:
     return image_name, image_tag
 
 
-def _print_logs_to_stdout(namespace: str, job_or_deployment_name: str) -> None:
+def _compute_optimal_job_timeout(batch_stages: List[BatchStageConfig]) -> int:
+    """Compute the optimal timeout for job monitoring.
+
+    :param namesapce: The target namespace for the job.
+    :param batch_stages: The desired configuration for the jobs.
+    :param returns: The optimal timeout (in seconds).
+    """
+    job_timeouts = [
+        max(1, stage.retries) * stage.max_completion_time for stage in batch_stages
+    ]
+    return int(max(job_timeouts) + TIMEOUT_GRACE_SECONDS)
+
+
+def _compute_optimal_deployment_timeout(
+    namespace: str, service_stages: List[ServiceStageConfig]
+) -> int:
+    """Compute the optimal timeout for deployment monitoring.
+
+    If a deployment is rolling-out for the first time, then wait for
+    twice the max configured startup time. If a deployment is updating,
+    then compute how long it will take to replace all pods using the
+    rolling update strategy.
+
+    The the max configured startup time has been floored at 60s, because
+    installing just Flask alone takes this long and it would be easy to
+    incorrectly estimate this.
+
+    :param namespace: The target namespace for the deployments.
+    :param service_stages: The desired configuration for the incoming
+        deployments.
+    :param returns: The optimal timeout (in seconds).
+    """
+    new_pod_rate = K8S_MAX_SURGE + K8S_MAX_UNAVAILABLE
+    deployment_timeouts = [
+        (
+            stage.max_startup_time
+            if not k8s.is_existing_deployment(namespace, stage.name)
+            else ceil(stage.replicas / new_pod_rate) * max(60, stage.max_startup_time)
+        )
+        for stage in service_stages
+    ]
+    return int(2 * max(deployment_timeouts) + TIMEOUT_GRACE_SECONDS)
+
+
+def _print_logs_to_stdout(
+    namespace: str, job_or_deployment_name: str, previous: bool = False
+) -> None:
     """Replay pod logs from a job or deployment to stdout.
 
     :param namespace: The namespace the job/deployment is in.
     :param job_or_deployment_name: The name of the pod or deployment.
+    :param previous: Return logs from previously crashed pod.
     """
     try:
         pod_name = k8s.get_latest_pod_name(namespace, job_or_deployment_name)
         if pod_name is not None:
-            pod_logs = k8s.get_pod_logs(namespace, pod_name)
+            pod_logs = k8s.get_pod_logs(namespace, pod_name, previous)
             print_pod_logs(pod_logs, f"logs for stage = {pod_name}")
         else:
             _log.warning(f"Cannot get logs for {job_or_deployment_name}")
